<?php

namespace Drupal\ai_provider_dreamstudio\Plugin\AiProvider;

use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\ai\Attribute\AiProvider;
use Drupal\ai\Base\AiProviderClientBase;
use Drupal\ai\Exception\AiResponseErrorException;
use Drupal\ai\OperationType\GenericType\ImageFile;
use Drupal\ai\OperationType\TextToImage\TextToImageInput;
use Drupal\ai\OperationType\TextToImage\TextToImageInterface;
use Drupal\ai\OperationType\TextToImage\TextToImageOutput;
use Drupal\ai_provider_dreamstudio\DreamStudio;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Yaml\Yaml;

/**
 * Plugin implementation of the 'dreamstudio' provider.
 */
#[AiProvider(
  id: 'dreamstudio',
  label: new TranslatableMarkup('DreamStudio'),
)]
class DreamStudioProvider extends AiProviderClientBase implements
  ContainerFactoryPluginInterface,
  TextToImageInterface {

  /**
   * The DreamStudio Client.
   *
   * @var \Drupal\dreamstudio\DreamStudio
   */
  protected $client;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The temporary files.
   *
   * @var array
   */
  protected $temporaryFiles = [];

  /**
   * Destructor.
   */
  public function __destruct() {
    foreach ($this->temporaryFiles as $file) {
      $file->delete();
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): static {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->client = $container->get('dreamstudio.api');
    $instance->entityTypeManager = $container->get('entity_type.manager');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function getConfiguredModels(?string $operation_type = NULL, array $capabilities = []): array {
    $models = [];
    if (is_null($operation_type) || $operation_type === 'text_to_image') {
      $models['stable-diffusion-xl-1024-v1-0'] = 'Stable Diffusion XL 1024 v1.0';
      $models['stable-diffusion-v1-6'] = 'Stable Diffusion v1.6';
      $models['stable-image-ultra'] = 'Stable Image Ultra';
      $models['stable-image-core'] = 'Stable Image Core';
      $models['sd3-large'] = 'Stable Diffusion 3 Large';
      $models['sd3-large-turbo'] = 'Stable Diffusion 3 Large Turbo';
      $models['sd3-medium'] = 'Stable Diffusion 3 Medium';
    }
    return $models;
  }

  /**
   * {@inheritdoc}
   */
  public function isUsable(?string $operation_type = NULL, array $capabilities = []): bool {
    // If its not configured, it is not usable.
    if (!$this->getConfig()->get('api_key')) {
      return FALSE;
    }
    // If its one of the bundles that Mistral supports its usable.
    if ($operation_type) {
      return in_array($operation_type, $this->getSupportedOperationTypes());
    }
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function getSupportedOperationTypes(): array {
    return [
      'text_to_image',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getConfig(): ImmutableConfig {
    return $this->configFactory->get('ai_provider_dreamstudio.settings');
  }

  /**
   * {@inheritdoc}
   */
  public function getApiDefinition(): array {
    // Load the configuration.
    $definition = Yaml::parseFile($this->moduleHandler->getModule('ai_provider_dreamstudio')->getPath() . '/definitions/api_defaults.yml');
    return $definition;
  }

  /**
   * {@inheritdoc}
   */
  public function getModelSettings(string $model_id, array $generalConfig = []): array {
    switch ($model_id) {
      case 'stable-diffusion-xl-1024-v1-0':
        unset($generalConfig['aspect_ratio']);
        unset($generalConfig['output_format']);
        $generalConfig['width']['default'] = 1536;
        $generalConfig['height']['default'] = 640;
        break;

      case 'stable-diffusion-v1-6':
        unset($generalConfig['aspect_ratio']);
        unset($generalConfig['output_format']);
        break;

      case 'stable-image-ultra':
        unset($generalConfig['width']);
        unset($generalConfig['height']);
        unset($generalConfig['style_preset']);
        unset($generalConfig['steps']);
        break;

      case 'sd3-large':
      case 'sd3-large-turbo':
      case 'sd3-medium':
        unset($generalConfig['width']);
        unset($generalConfig['height']);
        unset($generalConfig['style_preset']);
        unset($generalConfig['steps']);
        $generalConfig['output_format']['constraints']['options'] = [
          'jpeg',
          'png',
        ];
        break;

      case 'stable-image-core':
        unset($generalConfig['width']);
        unset($generalConfig['height']);
        break;
    }
    return $generalConfig;
  }

  /**
   * {@inheritdoc}
   */
  public function setAuthentication(mixed $authentication): void {
    // Set the new API key and reset the client.
    $this->client->setApiKey($authentication);
  }

  /**
   * {@inheritdoc}
   */
  public function textToImage(string|TextToImageInput $input, string $model_id, array $tags = []): TextToImageOutput {
    // Normalize the input if needed.
    if ($input instanceof TextToImageInput) {
      $input = $input->getText();
    }
    // Then send.
    $response = NULL;
    switch ($model_id) {
      case 'stable-diffusion-xl-1024-v1-0':
      case 'stable-diffusion-v1-6':
        if ($this->configuration['style_preset'] === '') {
          unset($this->configuration['style_preset']);
        }
        $response = $this->client->generateImage($input, $model_id, $this->configuration);
        break;

      case 'stable-image-ultra':
        $response = $this->client->generateImageUltra($input, $this->configuration);
        break;

      case 'sd3-large':
      case 'sd3-large-turbo':
      case 'sd3-medium':
        $config = $this->configuration + [
          'model' => $model_id,
        ];
        $response = $this->client->generateImageStableDiffusion3($input, $config);
        break;

      case 'stable-image-core':
        if ($this->configuration['style_preset'] === '') {
          unset($this->configuration['style_preset']);
        }
        $response = $this->client->generateImageStableCore($input, $this->configuration);
        break;
    }

    if (empty($response)) {
      throw new AiResponseErrorException('No image data found in the response.');
    }
    $filename = 'dreamstudio.png';
    $mimetype = 'image/png';
    if (isset($this->configuration['output_format']) && $this->configuration['output_format'] !== 'jpeg') {
      $filename = 'dreamstudio.jpg';
      $mimetype = 'image/jpeg';
    }
    if (isset($this->configuration['output_format']) && $this->configuration['output_format'] !== 'webp') {
      $filename = 'dreamstudio.webp';
      $mimetype = 'image/webp';
    }
    $images[] = new ImageFile($response, $mimetype, $filename);
    return new TextToImageOutput($images, $response, []);
  }

  /**
   * Gets the raw client.
   *
   * This is the client for inference.
   *
   * @return \Drupal\dreamstudio\DreamStudio
   *   The client.
   */
  public function getClient(): DreamStudio {
    return $this->client;
  }

}
